/****************************************************************
  Prolog-java interface
	by Naoyuki Tamura (tamura@kobe-u.ac.jp)
	   Dept. CS, Kobe University
****************************************************************/
import java.awt.*;
import java.awt.event.*;
import java.io.*;

public class PrologControlPanel extends Panel {
    private String prologCommand;
    private PrologServer prolog = null;
    private Button startButton;
    private Button nextButton;
    private Button abortButton;
    private Button quitButton;
    private TextField message;
    private int count = 0;
    private int state;
    public final int START  = 1;
    public final int WAIT_S = 2;
    public final int WAIT_R = 3;
    public final int NEXT   = 4;
    public final int DONE   = 5;
    public final int ABORT  = 6;
    public final int QUIT   = 7;
    public final int ERROR  = 8;

    public PrologControlPanel(String command) {
	prologCommand = command;
	// Create buttons and message-area
	startButton = new Button("Start");
	nextButton = new Button("Next");
	abortButton = new Button("Abort");
	quitButton = new Button("Quit");
	message = new TextField(20);
	message.setEditable(false);
	// Add ActionListeners
	startButton.addActionListener(new ActionListener() {
	    public void actionPerformed(ActionEvent e) { actionStart(); }
	});
	nextButton.addActionListener(new ActionListener() {
	    public void actionPerformed(ActionEvent e) { actionNext(); }
	});
	abortButton.addActionListener(new ActionListener() {
	    public void actionPerformed(ActionEvent e) { actionAbort(); }
	});
	quitButton.addActionListener(new ActionListener() {
	    public void actionPerformed(ActionEvent e) { actionQuit(); }
	});
	// Layout buttons and message-area
	Panel p = new Panel();
	p.add(startButton);
	p.add(nextButton);
	p.add(abortButton);
	p.add(quitButton);
	setLayout(new BorderLayout());
	add("Center", p);
	add("South", message);
	validate();
	setState(START);
    }

    public void close() {
	if (prolog == null)
	  return;
	prolog.close();
	prolog = null;
    }

    public synchronized void sendArgument(String arg) {
	if (arg == null)
	  return;
	if (state != WAIT_S || prolog == null)
	  return;
	try {
	    prolog.send(arg);
	    setState(WAIT_R);
	} catch (IOException e) {
	    close();
	    setState(ERROR);
	}
    }

    public synchronized String receiveResult() {
	String result = null;
	if (state != WAIT_R || prolog == null)
	  return null;
	try {
	    result = prolog.receive();
	} catch (IOException e) {
	    close();
	    setState(ERROR);
	    return null;
	}
	if (result.equals("DONE")) {
	    setState(DONE);
	} else {
	    count++;
	    setState(NEXT);
	}
	return result;
    }

    public synchronized int getState() {
	return state;
    }

    private synchronized void setState(int s) {
	switch (s) {
	  case START:
	    startButton.setEnabled(true);
	    startButton.requestFocus();
	    nextButton.setEnabled(false);
	    abortButton.setEnabled(false);
	    quitButton.setEnabled(true);
	    message.setText("");
	    break;
	  case WAIT_S:
	    startButton.setEnabled(false);
	    nextButton.setEnabled(false);
	    abortButton.setEnabled(true);
	    abortButton.requestFocus();
	    quitButton.setEnabled(true);
	    message.setText("Sending");
	    break;
	  case WAIT_R:
	    startButton.setEnabled(false);
	    nextButton.setEnabled(false);
	    abortButton.setEnabled(true);
	    abortButton.requestFocus();
	    quitButton.setEnabled(true);
	    message.setText("Receiving");
	    break;
	  case NEXT:
	    startButton.setEnabled(true);
	    nextButton.setEnabled(true);
	    nextButton.requestFocus();
	    abortButton.setEnabled(true);
	    quitButton.setEnabled(true);
	    message.setText("Count = " + count);
	    break;
	  case DONE:
	    startButton.setEnabled(true);
	    startButton.requestFocus();
	    nextButton.setEnabled(false);
	    abortButton.setEnabled(false);
	    quitButton.setEnabled(true);
	    message.setText("Count = " + count + " (no more)");
	    break;
	  case ABORT:
	    startButton.setEnabled(true);
	    startButton.requestFocus();
	    nextButton.setEnabled(false);
	    abortButton.setEnabled(false);
	    quitButton.setEnabled(true);
	    message.setText("Aborted");
	    break;
	  case QUIT:
	    startButton.setEnabled(false);
	    nextButton.setEnabled(false);
	    abortButton.setEnabled(false);
	    quitButton.setEnabled(false);
	    message.setText("Bye");
	    break;
	  case ERROR:
	    startButton.setEnabled(true);
	    nextButton.setEnabled(false);
	    abortButton.setEnabled(false);
	    quitButton.setEnabled(true);
	    quitButton.requestFocus();
	    message.setText("Error");
	    break;
	}
	state = s;
	repaint();
    }

    public synchronized void actionStart() {
	try {
	    if (prolog == null)
	      prolog = new PrologServer(prologCommand);
	    else
	      prolog.send("");
	    setState(WAIT_S);
	} catch (IOException e) {
	    close();
	    setState(ERROR);
	}
	count = 0;
    }

    public synchronized void actionNext() {
	if (prolog == null)
	  return;
	try {
	    prolog.send(";");
	    setState(WAIT_R);
	} catch (IOException e) {
	    close();
	    setState(ERROR);
	}
    }

    public void actionAbort() {
	close();
	setState(ABORT);
    }

    public void actionQuit() {
	close();
	setState(QUIT);
    }

    public void paint(Graphics g) {
	paintComponents(g);
    }

}


